{ config, lib, pkgs, ... }:
let
  inherit (config) xdg;
  shellAbbrs = lib.filterAttrs (_: val: builtins.typeOf val == "string") config.programs.fish.shellAbbrs;
in
with lib;
{
  # Set up abbreviations via the user's configuration file
  xdg.configFile."zsh-abbr/user-abbreviations".text = ''
    ${concatLines (mapAttrsToList (key: val: "abbr --quiet ${escapeShellArg key}=${escapeShellArg val}" ) shellAbbrs)}
  '';

  programs.zsh = {
    dotDir = ".config/zsh";
    enableCompletion = true;

    # These variables needs to be set before sourcing plugins
    initExtraFirst = ''
      # Load all the settings immediately on sourcing, so
      # they can be overridden later by other plugins
      export ZVM_INIT_MODE=sourcing
      export ZVM_LINE_INIT_MODE=$ZVM_MODE_INSERT

      export ZSH_AUTOSUGGEST_STRATEGY=(history completion)
      export ZSH_AUTOSUGGEST_BUFFER_MAX_SIZE=30
    '';

    # Ref: zshcompsys(1)
    # - https://github.com/hlissner/dotfiles/blob/master/config/zsh/completion.zsh
    # - https://thevaluable.dev/zsh-completion-guide-examples/
    completionInit = ''
      autoload -Uz compinit
      zmodload zsh/complist
      compinit -d ${escapeShellArg xdg.cacheHome}/zsh/zcompdump-"$ZSH_VERSION"

      # General completion menu config
      zstyle ':completion:*' menu select
      zstyle ':completion:*' group-name '''
      zstyle ':completion:*' rehash true
      zstyle ':completion:*' special-dirs true
      zstyle ':completion:*' verbose true

      # Cache completion list
      zstyle ':completion::complete:*' use-cache true
      zstyle ':completion::complete:*' cache-path ${escapeShellArg xdg.cacheHome}/zsh/zcompcache

      # Fuzzy matching
      zstyle ':completion:*' completer _complete _match _approximate
      zstyle ':completion:*:match:*' original only
      zstyle ':completion:*:approximate:*' max-errors 1 numeric

      # Use LS_COLORS for completion list
      # (LS_COLORS should be set before this is run)
      zstyle ':completion:*:default' list-colors ''${(s.:.)LS_COLORS}
      # Make color
      zstyle ':completion:*:corrections' format '%B-- %F{green}%d%f (errors: %F{red}%e%f) --%b'
      zstyle ':completion:*:descriptions' format '%B%F{cyan}[%d]%f%b'
      zstyle ':completion:*:messages' format '%F{yellow}-- %d --%f'
      zstyle ':completion:*:warnings' format '%B%F{red}-- No match for: %d --%f%b'

      # Also complete options for cd, chdir, pushd
      zstyle ':completion:*' complete-options true
      # Shorten directories' names when cd-ing
      zstyle ':completion:*:paths' path-completion true
      # Case insensitive
      zstyle ':completion:*' matcher-list 'm:{[:lower:][:upper:]}={[:upper:][:lower:]}' 'r:|=*' 'l:|=* r:|=*'
      # Expand slashes the UNIX way
      zstyle ':completion:*' squeeze-slashes true
      # Make the completion list scrollable
      zstyle ':completion:*:default' list-prompt '%S%M matches%s'
      # Don't complete unavailable commands.
      zstyle ':completion:*:functions' ignored-patterns '(_*|pre(cmd|exec))'
      # Array completion element sorting.
      zstyle ':completion:*:*:-subscript-:*' tag-order indexes parameters
    '';

    initExtra = ''
      # Set zsh options. See zshoptions(1)
      source ${./config.zsh}

      # Ensure these binary modules are available. See zshmodules(1)
      zmodload zsh/attr zsh/zle zsh/zpty

      # Load nice addons. See zshcontrib(1)
      autoload -Uz edit-command-line zcalc zmv zargs
      zle -N edit-command-line

      # Enable aliases of color names to corresponding escape sequences
      autoload -Uz colors && colors

      # Vi mode
      bindkey -v

      # zsh-autosuggestions
      bindkey -M viins '^F' autosuggest-accept
      # zsh-history-substring-search
      bindkey '^[[A' history-substring-search-up
      bindkey '^[[B' history-substring-search-down
      bindkey -M viins '^P' history-substring-search-up
      bindkey -M viins '^N' history-substring-search-down
      bindkey -M vicmd 'k' history-substring-search-up
      bindkey -M vicmd 'j' history-substring-search-down
      # Navigate the completion menu
      bindkey -M menuselect '^[[Z' reverse-menu-complete
      bindkey -M menuselect '^H' vi-backward-char
      bindkey -M menuselect '^J' vi-down-line-or-history
      bindkey -M menuselect '^K' vi-up-line-or-history
      bindkey -M menuselect '^L' vi-forward-char
      # Custom key bindings
      bindkey '^[e' edit-command-line
      bindkey -M viins '^A' beginning-of-line
      bindkey -M viins '^E' end-of-line

      # Cute bunnies are important
      ${escapeShellArg xdg.binHome}/bunny.sh
    '';

    history = {
      path = "${xdg.dataHome}/zsh/zsh_history";
      size = 12000;
      save = 10000;
      extended = true;
      expireDuplicatesFirst = true;
      ignorePatterns = [ "rm *" "pkill *" "kill *" ];
    };

    plugins = [
      {
        name = "zsh-vi-mode";
        src = pkgs.fetchFromGitHub {
          owner = "jeffreytse";
          repo = "zsh-vi-mode";
          rev = "ea1f58ab9b1f3eac50e2cde3e3bc612049ef683b";
          hash = "sha256-xbchXJTFWeABTwq6h4KWLh+EvydDrDzcY9AQVK65RS8=";
        };
      }
      {
        name = "fast-syntax-highlighting";
        src = pkgs.fetchFromGitHub {
          owner = "zdharma-continuum";
          repo = "fast-syntax-highlighting";
          rev = "cf318e06a9b7c9f2219d78f41b46fa6e06011fd9";
          hash = "sha256-RVX9ZSzjBW3LpFs2W86lKI6vtcvDWP6EPxzeTcRZua4=";
        };
      }
      {
        name = "zsh-history-substring-search";
        src = pkgs.fetchFromGitHub {
          owner = "zsh-users";
          repo = "zsh-history-substring-search";
          rev = "8dd05bfcc12b0cd1ee9ea64be725b3d9f713cf64";
          hash = "sha256-houujb1CrRTjhCc+dp3PRHALvres1YylgxXwjjK6VZA=";
        };
      }
      {
        name = "zsh-autosuggestions";
        src = pkgs.fetchFromGitHub {
          owner = "zsh-users";
          repo = "zsh-autosuggestions";
          rev = "c3d4e576c9c86eac62884bd47c01f6faed043fc5";
          hash = "sha256-B+Kz3B7d97CM/3ztpQyVkE6EfMipVF8Y4HJNfSRXHtU=";
        };
      }
      {
        name = "zsh-autopair";
        src = pkgs.fetchFromGitHub {
          owner = "hlissner";
          repo = "zsh-autopair";
          rev = "2ec3fd3c9b950c01dbffbb2a4d191e1d34b8c58a";
          hash = "sha256-Y7fkpvCOC/lC2CHYui+6vOdNO8dNHGrVYTGGNf9qgdg=";
        };
      }
      {
        name = "zsh-abbr";
        src = pkgs.fetchFromGitHub {
          owner = "olets";
          repo = "zsh-abbr";
          rev = "447708a39127552c92ee5b785d79dc5ce2bf475f";
          hash = "sha256-gEBGMVR1lMVKNPVuPjtdPkgOXI1MWO0EAtk7JRmS0Ok=";
        };
      }
    ];
  };
}
