{ lib, pkgs, ... }:
{
  imports = [
    ./users
    ./hardware.nix
  ];

  sops = {
    defaultSopsFile = ./secrets.json;
    defaultSopsFormat = "json";

    # NOTE: the ssh key needs to exist (and can be used to decrypt secrets) before running `nixos-install`
    age.sshKeyPaths = [ "/persist/etc/ssh/ssh_host_ed25519_key" ];

    # Another approach is to only use the master key to encrypt secrets on `nixos-install`.
    # - Copy the master Age key to the installer
    # - Run `nixos-install`
    # - Get the generated SSH host keys (see services.openssh.hostKeys below)
    # - Use ssh-to-age to get the public Age key, and use it to rotate the encrypted secrets (sops updatekeys)
    # - `nixos-rebuild switch` should work from now on (and the above configuration can be uncommented)
    # age.keyFile = "/var/lib/sops-nix/age-key.txt";
  };

  # This setting is solely used to ensure ZFS can import pools correctly.
  # Randomly generated with `head -c4 /dev/urandom | od -A none -t x4` (ensure it's unique per network).
  networking.hostId = "b5ce5013";

  # Can't live without Internet, can we?
  networking.networkmanager = {
    enable = true;
    dns = "systemd-resolved";
    wifi.backend = "iwd";
  };

  # Use an NTP client with NTS support (systemd-timesyncd is the default)
  # Currently, there are only 2 options: chrony and ntpd-rs
  services.ntpd-rs =
    let
      timeServers = [
        "time.cloudflare.com"
        "ntpmon.dcs1.biz"
        "nts.netnod.se"
        "ntp.zeitgitter.net"
        "virginia.time.system76.com"
        "ntp3.fau.de"
        "gps.ntp.br"
      ];
    in
    {
      enable = true;
      useNetworkingTimeServers = false;
      settings = {
        source = builtins.map
          (server: {
            mode = "nts";
            address = server;
          })
          timeServers;
      };
    };

  # Basic protection is good
  networking.nftables.enable = true;
  networking.nftables.flushRuleset = true;
  networking.firewall = {
    enable = true;
    filterForward = true;
    pingLimit = "10/second burst 5 packets";

    # TODO: allow DNS queries for libvirt and incus bridge interfaces
    # Ref: https://github.com/NixOS/nixpkgs/issues/263359
  };

  # Trying out the hot new sudo "rewritten in Rust"
  security.sudo-rs.enable = true;

  # Not as horrifying as SELinux, and way less effective
  security.apparmor.enable = true;
  services.dbus.apparmor = "enabled";

  # DNS-over-TLS out of the box. Yay.
  services.resolved =
    let
      # Ref: https://dnsprivacy.org/public_resolvers/#dns-over-tls-dot
      dnsServers = [
        "9.9.9.9#dns.quad9.net"
        "149.112.112.112#dns.quad9.net"
        "2620:fe::fe#dns.quad9.net"
        "2620:fe::9#dns.quad9.net"
        "1.1.1.1#cloudflare-dns.com"
        "1.0.0.1#cloudflare-dns.com"
        "2606:4700:4700::1111#cloudflare-dns.com"
        "2606:4700:4700::1001#cloudflare-dns.com"
      ];
    in
    {
      enable = true;
      dnssec = "true";
      dnsovertls = "true";
      fallbackDns = lib.mkForce [ ]; # disable compiled-in fallback DNS list
      extraConfig = ''
        DNS=${builtins.concatStringsSep " " dnsServers}
      '';
    };

  # Make sshd use host keys stored in the persisted directory
  services.openssh = {
    # enable = true;
    hostKeys = [
      { path = "/persist/etc/ssh_host_ed25519_key"; type = "ed25519"; }
      { path = "/persist/etc/ssh_host_rsa_key"; type = "rsa"; bits = 4096; }
    ];
  };

  # Rotate logs
  services.logrotate = {
    enable = true;

    # Persist the state file where it can survive reboots
    extraArgs = [ "--state" "/persist/var/lib/logrotate.status" ];
  };

  # Keep the disks in good health
  services.zfs.autoScrub = { enable = true; interval = "monthly"; };
  services.btrfs.autoScrub = {
    enable = true;
    interval = "monthly";
    fileSystems = [ "/media" ];
  };
  services.sanoid = {
    enable = true;
    interval = "hourly";
    templates.production = {
      autoprune = true;
      autosnap = true;
      yearly = 0;
      monthly = 1;
      weekly = 2;
      daily = 4;
      hourly = 8;
    };
    datasets."zroot/home" = {
      useTemplate = [ "production" ];
      recursive = true;
      processChildrenOnly = true;
    };
  };
  services.btrbk.instances.btrbk = {
    onCalendar = "hourly";
    settings = {
      timestamp_format = "long";
      lockfile = "/var/lib/btrbk/btrbk.lock";
      cache_dir = "/var/cache/btrbk";
      btrfs_commit_delete = "yes";

      subvolume."/media" = {
        snapshot_create = "onchange";
        snapshot_preserve_min = "latest";
        snapshot_preserve = "16h 8d 4w 2m";
        snapshot_dir = "/media/.snapshots";
      };
    };
  };

  # Keep the firmware up-to-date
  services.fwupd.enable = true;

  # Such a crime, using Guix on a NixOS system
  services.guix = { enable = true; nrBuildUsers = 8; };

  # TODO: enable libvirt and incus services (use modular libvirt daemons)
  # Some notes:
  # - https://nixos.wiki/wiki/Libvirt
  # - https://github.com/NixOS/nixpkgs/pull/258379
  # virtualisation.libvirtd = {
  #   enable = true;
  #   qemu = {
  #     runAsRoot = false;
  #     swtpm.enable = true;
  #   };
  # };

  # TODO: enable auditd, usbguard

  # Let everyone use flatpak
  services.flatpak.enable = true;

  # Vietnamese and Japanese
  # TODO: add configuration for the input methods
  i18n.inputMethod.enabled = "fcitx5";
  i18n.inputMethod.fcitx5 = {
    waylandFrontend = true;
    addons = with pkgs; [
      fcitx5-bamboo
      fcitx5-mozc
      kdePackages.fcitx5-with-addons
    ];
  };

  # KDE is easy to use. I'll have Sway later.
  # QT platformtheme and cursor theme are managed per-user inside home-manager.
  # Ref: https://nixos.wiki/wiki/KDE
  services.xserver.enable = true;
  services.displayManager.defaultSession = "plasma";
  services.displayManager.sddm = {
    enable = true;
    enableHidpi = true;
    wayland.enable = true;
    theme = "breeze";
  };
  services.desktopManager.plasma6.enable = true;
  environment.plasma6.excludePackages = with pkgs.kdePackages; [
    plasma-browser-integration
    konsole
  ];
  xdg.portal = {
    xdgOpenUsePortal = true;
    extraPortals = [ pkgs.xdg-desktop-portal-gtk ];
  };

  # Sound
  security.rtkit.enable = true;
  services.pipewire = {
    enable = true;
    pulse.enable = true;
    jack.enable = true;
  };

  # Most valuable directories (.e.g /home, /gnu, /nix) are persisted using ZFS datasets.
  # Impermanence is used only to deal with state files.
  environment.persistence."/persist" = {
    hideMounts = true;
    files = [
      "/etc/machine-id"
    ];
    directories = [
      "/etc/NetworkManager/system-connections"
      "/var/cache/btrbk"
      "/var/lib/libvirt"
      "/var/lib/nixos"
      "/var/lib/systemd"
      "/var/log"
    ];
  };

  # Ref: https://nixos.wiki/wiki/Fonts
  fonts = {
    enableDefaultPackages = true;
    packages = with pkgs; [
      sarasa-gothic
      iosevka-bin
      (iosevka-bin.override { variant = "Curly"; })
      (iosevka-bin.override { variant = "CurlySlab"; })
      (iosevka-bin.override { variant = "Slab"; })
      (iosevka-bin.override { variant = "Aile"; })
      (iosevka-bin.override { variant = "Etoile"; })
      (nerdfonts.override { fonts = [ "IosevkaTerm" ]; })
    ];
    fontconfig.defaultFonts = {
      serif = [
        "DejaVu Serif"
        "Iosevka Etoile"
        "Sarasa Fixed Slab J"
      ];
      sansSerif = [
        "DejaVu Sans"
        "Iosevka Aile"
        "Sarasa Fixed J"
      ];
      monospace = [
        "DejaVu Sans Mono"
        "Iosevka"
        "Sarasa Mono J"
      ];
    };
  };

  # Set your time zone.
  time.timeZone = "Asia/Ho_Chi_Minh";

  # NOTE: never ever change this value after the installation is finished
  system.stateVersion = "24.05";
}
