{ config, pkgs, ... }: {
  boot.initrd.availableKernelModules = [ "nvme" "xhci_pci" "ahci" "usb_storage" "usbhid" "sd_mod" ];
  boot.initrd.kernelModules = [ "amdgpu" ];
  boot.initrd.supportedFilesystems = [ "zfs" ];
  boot.initrd.systemd.enable = true; # allow plymouth to prompt for LUKS password

  boot.kernelModules = [ "kvm-amd" ];
  boot.kernelParams = [ "amd_pstate=active" ];
  boot.kernelPackages = config.boot.zfs.package.latestCompatibleLinuxPackages;

  boot.loader.efi.canTouchEfiVariables = true;
  boot.loader.timeout = 10;
  boot.loader.grub = {
    inherit (config.boot.plymouth) font;
    enable = true;
    zfsSupport = true;
    efiSupport = true;
    configurationLimit = 10;
    fontSize = 24;
    mirroredBoots = [
      { devices = [ "nodev" ]; path = "/boot"; }
    ];
  };

  # Being fancy. Other people have already taken care of all the tough setting-up parts :)
  boot.plymouth = {
    enable = true;
    themePackages = [ pkgs.adi1090x-plymouth-themes ];
    theme = "rings_2";
  };

  # Terminus is a nice console font set
  console.earlySetup = true;
  console.font = "${pkgs.terminus_font}/share/consolefonts/ter-h32b.psf.gz";

  # Necessary ~~evils~~ binary blobs
  # Ref: https://github.com/NixOS/nixos-hardware/tree/master/common/{cpu,gpu}/amd
  hardware.cpu.amd.updateMicrocode = true;
  hardware.enableRedistributableFirmware = true;

  # OpenGL, for wayland support in general
  hardware.opengl = {
    enable = true;
    driSupport = true;
    extraPackages = with pkgs.rocmPackages; [ clr clr.icd ];
  };

  # Bluetooth is meh, but I might need it sometimes
  hardware.bluetooth = {
    enable = true;
    powerOnBoot = false;
  };

  # For XP-Pen Artist 16 Pro (1st Gen) is not yet supported, though
  hardware.opentabletdriver.enable = true;

  # For the split keyboard
  hardware.keyboard.qmk.enable = true;

  disko.devices = {
    nodev."/" = {
      fsType = "tmpfs";
      mountOptions = [ "defaults" "size=2G" "mode=755" ];
    };

    disk = {
      sda = {
        device = "/dev/disk/by-id/wwn-0x50014ee213bce65e";
        type = "disk";
        content = {
          type = "gpt";
          partitions.media = {
            size = "100%";
            content = {
              type = "btrfs";
              extraArgs = [
                "--force"
                "--checksum blake2" # more "secure" than the default crc32c, still relatively fast
                "--features block-group-tree" # NOTE: put here until it's the default
              ];
              subvolumes =
                let
                  btrfsMountOptions = [
                    "nofail" # this is a pluggable external drive
                    "noatime"
                    "nodev"
                    "noexec"
                    "nosuid"
                    # "autodefrag"
                    # "user_subvol_rm_allowed"
                    "rescue=usebackuproot"
                    "compress-force=zstd:9"
                  ];
                in
                {
                  "@root" = {
                    mountpoint = "/media";
                    mountOptions = btrfsMountOptions;
                  };
                  "@snapshots" = {
                    mountpoint = "/media/.snapshots";
                    mountOptions = btrfsMountOptions;
                  };
                };
            };
          };
        };
      };

      nvme0n1 = {
        device = "/dev/disk/by-id/nvme-eui.002538423140a12b";
        type = "disk";
        content = {
          type = "gpt";
          partitions = {
            ESP = {
              size = "1G";
              type = "EF00";
              content = {
                type = "filesystem";
                format = "vfat";
                mountpoint = "/boot";
                mountOptions = [ "umask=0077" ];
              };
            };
            luks = {
              size = "100%";
              content = {
                type = "luks";
                name = "crypted";
                askPassword = true;
                content = {
                  type = "zfs";
                  pool = "zroot";
                };
              };
            };
          };
        };
      };
    };

    zpool.zroot = {
      type = "zpool";
      rootFsOptions = {
        # Sensible defaults
        acltype = "posixacl";
        canmount = "off";
        compression = "zstd-9";
        dnodesize = "auto";
        normalization = "formD";
        relatime = "on";
        xattr = "sa";
        mountpoint = "none";

        # Lower write frequency
        logbias = "throughput";
        recordsize = "16K";
      };
      options = {
        ashift = "12";
        autotrim = "on";
      };
      datasets = {
        # Over-provisioning
        reserved = {
          type = "zfs_fs";
          options = {
            mountpoint = "none";

            # NOTE: `refreservation` is preferred here over `reservation`, as it accounts for even snapshots. See zfsprops(7).
            # Ref: https://nex7.blogspot.com/2013/03/reservation-ref-reservation-explanation.html
            refreservation = "75G";
          };
        };
        nix = {
          type = "zfs_fs";
          mountpoint = "/nix";
          options = {
            devices = "off";
            atime = "off";
            mountpoint = "/nix";
          };
        };
        gnu = {
          type = "zfs_fs";
          mountpoint = "/gnu";
          options = {
            devices = "off";
            atime = "off";
            mountpoint = "/gnu";
          };
        };
        persist = {
          type = "zfs_fs";
          mountpoint = "/persist";
          options = {
            setuid = "off";
            exec = "off";
            devices = "off";
            mountpoint = "/persist";
          };
        };
        home = {
          type = "zfs_fs";
          options = {
            mountpoint = "/home";
            canmount = "off";
          };
        };
        "home/follie" = {
          type = "zfs_fs";
          mountpoint = "/home/follie";
          options.mountpoint = "/home/follie";
        };
        "home/gema" = {
          type = "zfs_fs";
          mountpoint = "/home/gema";
          options.mountpoint = "/home/gema";
        };
      };
    };
  };

  # Make the filesystem available early (disko doesn't support this option yet)
  # Ref: https://github.com/nix-community/disko/issues/192
  fileSystems."/persist".neededForBoot = true;

  nixpkgs.hostPlatform = "x86_64-linux";
}
